package com.javacodegeeks.aws.tasks;

import com.amazonaws.services.dynamodbv2.AmazonDynamoDB;
import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBMapper;
import com.amazonaws.services.dynamodbv2.document.*;
import com.amazonaws.services.dynamodbv2.document.spec.QuerySpec;
import com.amazonaws.services.dynamodbv2.document.utils.ValueMap;
import com.amazonaws.services.dynamodbv2.model.*;
import com.javacodegeeks.aws.model.TodoItem;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class QueriesTodoList implements Task {
    @Override
    public void run(String[] args, AmazonDynamoDB dynamodb) {
        createTodoListTable(dynamodb);
        insertSampleData(dynamodb);
        queryData(dynamodb);
        scanData(dynamodb);
        scanDataWithFilter(dynamodb);
        scanParallel(dynamodb);
    }

    private void scanParallel(final AmazonDynamoDB dynamodb) {
        final int numberOfSegments = 3;
        ExecutorService executorService = Executors.newFixedThreadPool(numberOfSegments);
        for (int i=0; i < numberOfSegments; i++) {
            final int segment = i;
            executorService.submit(new Runnable() {
                @Override
                public void run() {
                    Map<String, AttributeValue> lastEvaluatedKey = null;
                    while(true) {
                        ScanRequest scanRequest = new ScanRequest()
                                .withTableName("TodoList")
                                .withSegment(segment)
                                .withTotalSegments(numberOfSegments)
                                .withExclusiveStartKey(lastEvaluatedKey);
                        ScanResult scan = dynamodb.scan(scanRequest);
                        printScanResult(scan, segment);
                        lastEvaluatedKey = scan.getLastEvaluatedKey();
                        if (lastEvaluatedKey == null) {
                            break;
                        }
                    }
                }
            });
        }
        executorService.shutdown();
        try {
            executorService.awaitTermination(1, TimeUnit.HOURS);
        } catch (InterruptedException e) {
            executorService.shutdownNow();
        }
    }

    private void scanData(AmazonDynamoDB dynamodb) {
        ScanRequest scanRequest = new ScanRequest()
                .withTableName("TodoList").withLimit(100);
        ScanResult scan = dynamodb.scan(scanRequest);
        printScanResult(scan, -1);
    }

    private void printScanResult(ScanResult scan, int thread) {
        for (Map<String, AttributeValue> item : scan.getItems()){
            StringBuilder sb = new StringBuilder();
            for (Map.Entry entry : item.entrySet()) {
                if (sb.length() > 0) {
                    sb.append("; ");
                }
                sb.append(entry.getKey()).append("=").append(entry.getValue());
            }
            if (thread >= 0) {
                sb.append(" thread=").append(thread);
            }
            System.out.println(sb.toString());
        }
    }

    private void scanDataWithFilter(AmazonDynamoDB dynamodb) {
        Map<String, AttributeValue> expressionAttributeValues = new HashMap<>();
        expressionAttributeValues.put(":category", new AttributeValue().withS("Job"));

        ScanRequest scanRequest = new ScanRequest()
                .withTableName("TodoList")
                .withFilterExpression("Category = :category")
                .withExpressionAttributeValues(expressionAttributeValues)
                .withProjectionExpression("Category, Stat, description")
                .withLimit(10);
        ScanResult scan = dynamodb.scan(scanRequest);

        printScanResult(scan, -1);
    }

    private void queryData(AmazonDynamoDB dynamodb) {
        DynamoDB dynamoDB = new DynamoDB(dynamodb);
        Table table = dynamoDB.getTable("TodoList");

        QuerySpec spec = new QuerySpec()
                .withKeyConditionExpression("Category = :category")
                .withValueMap(new ValueMap()
                        .withString(":category", "Job"))
                .withMaxPageSize(10);

        ItemCollection<QueryOutcome> items = table.query(spec);
        printResult(items);

        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
        spec = new QuerySpec()
                .withKeyConditionExpression("Category = :category and CreationDate >= :creationDate")
                .withFilterExpression("Stat = :status")
                .withValueMap(new ValueMap()
                        .withString(":category", "House")
                        .withString(":creationDate", dtf.format(LocalDateTime.of(2017, 5, 17, 10, 0, 0)))
                    .withString(":status", "OPEN"))
                .withMaxPageSize(10);

        items = table.query(spec);
        printResult(items);
    }

    private void printResult(ItemCollection<QueryOutcome> items) {
        int pageNumber = 0;
        for (Page<Item, QueryOutcome> page : items.pages()) {
            pageNumber++;
            System.out.println("Page: " + pageNumber + " #####");
            for (Item item : page) {
                System.out.println(item.toJSONPretty());
            }
        }
    }

    private void insertSampleData(AmazonDynamoDB dynamodb) {
        DynamoDBMapper mapper = new DynamoDBMapper(dynamodb);
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("House")
                .withCreationDate(LocalDateTime.of(2017, 3, 1, 18, 0, 0))
                .withDescription("Paint living room")
                .withStatus("DONE").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("House")
                .withCreationDate(LocalDateTime.of(2017, 5, 17, 10, 0, 0))
                .withDescription("Repair kitchen sink")
                .withStatus("OPEN").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("House")
                .withCreationDate(LocalDateTime.of(2017, 9, 13, 9, 0, 0))
                .withDescription("Clean up cellar")
                .withStatus("DONE").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("House")
                .withCreationDate(LocalDateTime.of(2017, 9, 14, 9, 0, 0))
                .withDescription("Repair garage")
                .withStatus("OPEN").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("Job")
                .withCreationDate(LocalDateTime.of(2017, 10, 12, 9, 0, 0))
                .withDescription("Prepare meeting")
                .withStatus("OPEN").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("Job")
                .withCreationDate(LocalDateTime.of(2017, 10, 13, 9, 0, 0))
                .withDescription("Learn new programming language")
                .withStatus("OPEN").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("Finance")
                .withCreationDate(LocalDateTime.of(2017, 9, 13, 19, 0, 0))
                .withDescription("Sell stocks")
                .withStatus("OPEN").build());
        mapper.save(new TodoItem.TodoItemBuilder()
                .withCategory("Finance")
                .withCreationDate(LocalDateTime.of(2017, 8, 12, 7, 0, 0))
                .withDescription("Check balance")
                .withStatus("OPEN").build());
    }

    private void createTodoListTable(AmazonDynamoDB dynamodb) {
        ListTablesResult listTablesResult = dynamodb.listTables();
        List<String> tableNames = listTablesResult.getTableNames();
        for (String tableName : tableNames) {
            if ("TodoList".equals(tableName)) {
                DeleteTableRequest deleteTableReq = new DeleteTableRequest().withTableName(tableName);
                dynamodb.deleteTable(deleteTableReq);
                System.out.println("Deleted table " + tableName);
            }
        }

        ArrayList<AttributeDefinition> attributeDefinitions = new ArrayList<>();
        attributeDefinitions.add(new AttributeDefinition()
                .withAttributeName("Category").withAttributeType("S"));
        attributeDefinitions.add(new AttributeDefinition()
                .withAttributeName("CreationDate").withAttributeType("S"));

        ArrayList<KeySchemaElement> keySchema = new ArrayList<>();
        keySchema.add(new KeySchemaElement()
                .withAttributeName("Category").withKeyType(KeyType.HASH));
        keySchema.add(new KeySchemaElement()
                .withAttributeName("CreationDate").withKeyType(KeyType.RANGE));

        CreateTableRequest createTableReq = new CreateTableRequest()
                .withTableName("TodoList")
                .withAttributeDefinitions(attributeDefinitions)
                .withKeySchema(keySchema)
                .withProvisionedThroughput(new ProvisionedThroughput()
                        .withReadCapacityUnits(10L)
                        .withWriteCapacityUnits(5L));
        CreateTableResult result = dynamodb.createTable(createTableReq);
        System.out.println("Created table " + result.toString());
    }
}
